// InterceptorProtocol.swift
//
// Copyright © 2018-2021 Vasilis Panagiotopoulos. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of
// this software and associated documentation files (the "Software"), to deal in the
// Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
// and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies
// or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FIESS FOR A PARTICULAR
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import Foundation

/// This will be used in interceptor callback as an action to inteceptors chain.
public enum InterceptionAction {
    /// Continue with the next interceptor or final callbacks if there is no other interceptor in chain.
    case `continue`
    /// Retry the request
    /// - Parameters
    ///     - delay: The delay between retries in seconds. Pass nil value for no delay.
    case retry(delay: TimeInterval?)
}

/// Use this protocol to create interceptors that can be passed to Configuration instances.
/// Every class which implements this protocol will intercept between request completion and callbacks.
public protocol InterceptorProtocol {
    /// This function is called when a request is failed.
    ///   - parameters:
    ///     - responseData: The response data of request.
    ///     - error: The TNError provided by TermiNetwork on request failure, otherwise nil value is passed.
    ///     - request: The Request object.
    func requestFinished(responseData data: Data?,
                         error: TNError?,
                         request: Request,
                         proceed: @escaping (InterceptionAction) -> Void)

    /// Default initializer
    init()
}
